/*
  $Id: universe.c,v 1.43 1997/06/09 10:46:00 hannappe Exp $
  Modularized driver for Tundra semiconductor Universe PCI/VME Bridge
 */
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/delay.h>

#include <linux/types.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>
#include <linux/malloc.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>

#include <linux/bios32.h>
#include <linux/pci.h>

#include <asm/io.h>
#include <asm/semaphore.h>
#include <asm/bitops.h>
#include <asm/segment.h>
#include <asm/system.h>

#if LINUX_VERSION_CODE < 0x020100
#define ioremap vremap
#define iounmap vfree
#endif


#define UNIVERSE_REGISTERS_ARE_MEMORY_MAPPED
#ifdef UNIVERSE_REGISTERS_ARE_MEMORY_MAPPED
#define read_register(offset) readl(UCSR_base_addr + offset)
#define write_register(value,offset) writel(value,UCSR_base_addr + offset)
#define read_register_word(offset) readw(UCSR_base_addr + offset)
#define write_register_word(value,offset) writew(value,UCSR_base_addr + offset)
#else
#define read_register(offset) inl(UCSR_base_addr + offset)
#define write_register(value,offset) outl(value,UCSR_base_addr + offset)
#define read_register_word(offset) inw(UCSR_base_addr + offset)
#define write_register_word(value,offset) outw(value,UCSR_base_addr + offset)
#endif

#include "universe.h"
#include "vme_io.h"

#ifdef UNIVERSE_REGISTERS_ARE_MEMORY_MAPPED
static unsigned char *UCSR_base_addr = NULL;
#else
static unsigned int UCSR_base_addr = 0;
#endif

#define UNIVERSE_MAJOR (61)
#define UNIVERSE_MINOR_STAT (0)
#define UNIVERSE_MINOR_CTL (1)
#define UNIVERSE_MINOR_A24A16_USER (2)
#define UNIVERSE_MINOR_A24A16_SUPER (3)
#define UNIVERSE_MINOR_A32_0 (4)
#define UNIVERSE_MINOR_A32_1 (5)
#define UNIVERSE_MINOR_A32_2 (6)
#define UNIVERSE_MINOR_A32_3 (7)
#define UNIVERSE_MINOR_VMEIRQ (11)

#define UNIVERSE_STATUS_BUFFER_SIZE 4096


struct interrupt_counters {
  unsigned int acfail;
  unsigned int sysfail;
  unsigned int sw_int;
  unsigned int sw_iack;
  unsigned int verr;
  unsigned int lerr;
  unsigned int dma;
  unsigned int virq[7];
  unsigned int vown;
};

/* data type for the private_data field of the file type,
   see fs.h */
struct image_desc_type {
  char *kernel_addr;	    /* virt addr of mapped image in kernel space */
  size_t map_size;          /* size of the map at kernel_addr */
  unsigned long mmap_offset;/* offset from start of PCI slave image to
			       the address which is to be remapped at mmap()
			       time */
  unsigned long phys_start; /* phys start addr of PCI slave image */
  unsigned long phys_end;   /* phys end addr of PCI slave image */
  unsigned long vme_start;  /* start addr of image on the VME side */
  size_t dma_threshold;
  size_t dma_irq_threshold;
  unsigned long dctl;       /* value of DCLT-register if DMA is to be done */
};

static struct image_desc_type a32_img[4] = {{0},{0},{0},{0}};

#define UNIVERSE_A32_BASE_START (0x20000000)
#define UNIVERSE_A32_SPACING    (0x10000000)
#define UNIVERSE_A32_MAX_MAP_SIZE (64*1024*1024)
#define UNIVERSE_A32_DEFAULT_MAP_SIZE (16*1024*1024)

#define SLSI_A24D32_OFFSET 0
#define SLSI_A24D16_OFFSET 16*1024*1024
#define SLSI_A24D32_S_OFFSET 2*16*1024*1024
#define SLSI_A24D16_S_OFFSET 3*16*1024*1024
#define SLSI_A16_OFFSET (0x00ff0000)



/* address of the special image for A16/A24 access in kernel space */
static char *Slsi_kernel_addr = NULL;


static struct interrupt_counters Interrupt_counters = {0,0,
						       0,0,
						       0,0,0,
						       {0,0,0,0,0,0,0},
						       0};
static char *Status_buffer = NULL;
static size_t Status_pointer = 0;
static size_t Status_length = 0;

static int opens_of_ctl_device = 0;
static int opens_of_slsi_image = 0;

struct irq_device_type {
  int busy;
  int level;
  int vector;
  struct wait_queue *irqwait;
  int number_of_irqs_since_last_read;
  int statid;			/* status/id register content */
};
static struct irq_device_type irq_device[7][256] = {{{0}}};
static struct irq_device_type unassigned_irq_device = {0};

static struct semaphore DMA_available = MUTEX;
static size_t DMA_max_buffer_size = 64*1024;
static size_t _DMA_threshold = 34;
static size_t _DMA_irq_threshold = 256;
static struct wait_queue *DMA_wait = NULL;


static int universe_lseek(struct inode *inode, struct file *file,
		       off_t offset, int origin);
static int universe_read(struct inode *inode, struct file *file,
		      char *buf, int count);
static int universe_write(struct inode * inode, struct file *file,
		       const char * buf, int count);
static int universe_select(struct inode *inode, struct file *file,
			int sel_type, select_table *wait);
static int universe_ioctl(struct inode *inode, struct file *file,
			  unsigned int cmd, unsigned long arg);
static int universe_mmap(struct inode *inode, struct file *file,
			 struct vm_area_struct *vma);
static int universe_open(struct inode *inode, struct file *file);
static void universe_release(struct inode *inode, struct file *file);

static void universe_interrupt(int irq, void *dev_id, struct pt_regs *regs);


static int put_status (char *s);
static int put_status_int (unsigned int val, int radix);
static void init_status(void);
static int read_status (char *buf, int count);




static struct file_operations universe_fops = {
	universe_lseek,
	universe_read,
	universe_write,
	NULL,		/* No readdir */
	universe_select,
	universe_ioctl,
	universe_mmap,
	universe_open,
	universe_release
};

static inline int __get_order(unsigned long size)
{
	int order;

	size = (size-1) >> (PAGE_SHIFT-1);
	order = -1;
	do {
		size >>= 1;
		order++;
	} while (size);
	return order;
}


static int 
universe_lseek(struct inode *inode, struct file *file, off_t offset, int origin)
{
  unsigned int minor = MINOR(inode->i_rdev);
  long tmp = -1;

  switch (minor) {
  case UNIVERSE_MINOR_CTL:
  case UNIVERSE_MINOR_STAT:
    return -EINVAL;
    break;
  case UNIVERSE_MINOR_A24A16_USER:
  case UNIVERSE_MINOR_A24A16_SUPER:
  case UNIVERSE_MINOR_A32_0:
  case UNIVERSE_MINOR_A32_1:
  case UNIVERSE_MINOR_A32_2:
  case UNIVERSE_MINOR_A32_3:
    {
      struct image_desc_type *p;
      p = file->private_data;
      if (!p)
	{
	  printk("universe: missing private data in file struct\n");
	  return -EINVAL;
	}
      switch (origin) {
      case 0: tmp = offset; break;
      case 1: tmp = file->f_pos + offset; break;
      case 2: tmp = p->map_size + offset; break;
      }
      if (tmp < 0) return -EINVAL;
      if (tmp > p->map_size) return -EINVAL;
      file->f_pos = tmp;
      file->f_reada =0;
#ifdef UNIVERSE_DEBUG
      printk("universe: lseeked to %06lx\n",tmp);
#endif
      return(file->f_pos);
    }
  default:
    return -ENOSYS;
  }
  return -ENOSYS;
}

static int 
read_bytes(unsigned char* user_buffer,
	   unsigned char* vme_addr,
	   int count)
{
  int i;
  unsigned short pci_csr;
  for (i=0; i<count; i++)
    {
      unsigned char value;
      cli(); /* disable interrupts while reading value and register */ 
      write_register_word(PCI_STATUS_SIG_TARGET_ABORT, PCI_STATUS);
      value = readb(vme_addr);
      pci_csr = read_register_word(PCI_STATUS);
      sti(); /* re-enable interrupts, the read with check is done */

      if (pci_csr & PCI_STATUS_SIG_TARGET_ABORT)
	{
	  pci_csr = PCI_STATUS_SIG_TARGET_ABORT; /* writing a 1 clears... */
	  write_register_word(pci_csr, PCI_STATUS);
	  return(-EFAULT);
	}
      else
	{
	  put_user(value,user_buffer);
	  vme_addr++;
	  user_buffer++;
	}
    }
  return(count * sizeof(unsigned char));
} 
static int 
read_words(unsigned short* user_buffer,
	   unsigned short* vme_addr,
	   int count)
{
  int i;
  unsigned short pci_csr;
  for (i=0; i<count; i++)
    {
      unsigned short value;
      cli(); /* disable interrupts while reading value and register */ 
      write_register_word(PCI_STATUS_SIG_TARGET_ABORT, PCI_STATUS);
      value = readw(vme_addr);
      pci_csr = read_register_word(PCI_STATUS);
      sti(); /* re-enable interrupts, the read with check is done */

      if (pci_csr & PCI_STATUS_SIG_TARGET_ABORT)
	{
	  pci_csr = PCI_STATUS_SIG_TARGET_ABORT; /* writing a 1 clears... */
	  write_register_word(pci_csr, PCI_STATUS);
	  return(-EFAULT);
	}
      else
	{
	  put_user(value,user_buffer);
	  vme_addr++;
	  user_buffer++;
	}
    }
  return(count * sizeof(unsigned short));
} 
static int 
read_ints(unsigned int* user_buffer,
	  unsigned int* vme_addr,
	  int count)
{
  int i;
  unsigned short pci_csr;
  for (i=0; i<count; i++)
    {
      unsigned int value;
      cli(); /* disable interrupts while reading value and register */ 
      write_register_word(PCI_STATUS_SIG_TARGET_ABORT, PCI_STATUS);
      value = readl(vme_addr);
      pci_csr = read_register_word(PCI_STATUS);
      sti(); /* re-enable interrupts, the read with check is done */

      if (pci_csr & PCI_STATUS_SIG_TARGET_ABORT)
	{
	  pci_csr = PCI_STATUS_SIG_TARGET_ABORT; /* writing a 1 clears... */
	  write_register_word(pci_csr, PCI_STATUS);
	  return(-EFAULT);
	}
      else
	{
	  put_user(value,user_buffer);
	  vme_addr++;
	  user_buffer++;
	}
    }
  return(count*sizeof(unsigned int));
} 






static void
dma_timeout(unsigned long dummy)
{
  printk("universe: DMA timeout\n");
  wake_up_interruptible(&DMA_wait);
}

static inline unsigned int 
wait_for_dma_done(int use_irq)
{
  /* note: what happens if the sleep is interrupted and we do not
     return here? Then something must be done about
     both the DMA buffer and the DMA semaphore.
     */
  unsigned int dgcs;
  struct timer_list dma_timer;
  if (use_irq)
    {
      init_timer(&dma_timer);
      dma_timer.function = dma_timeout;
      dma_timer.data = 0;
      dma_timer.expires = jiffies + HZ/50;
      /* note: the correct timeout would be something like 
	 bytes_read_just_now*2/(6652474*datawidth_in_bytes*HZ),
	 but that gives always 0, because the maximum DMA buffer
	 size is just 64K and that is in any case handled after
	 1/100 second.
	 
	 The timeout is needed, because it is possible that
	 the DMA transfer finishes just between the read_register(UNIV_DGCS)
	 and the interruptible_sleep_on(&DMA_wait), so that the wait never
	 ends because the irq does not come, since it came just before the
	 wait queue was initialized.
	 So, with the timeout, we do really stop waiting, but possibly
	 a bit late, i.e. after the next timer interrupt or so.
	 */
      dgcs = read_register(UNIV_DGCS);
      if (dgcs & UNIV_BM_DGCS_ACT)
	{
	  add_timer(&dma_timer);
	  interruptible_sleep_on(&DMA_wait);
	}
      del_timer(&dma_timer);
      /* the dma should have finished by now,
	 just read DGCS again and make real sure that it has */
    }
  {
    int i;
    for (i=0; i<10000; i++)
      {
	dgcs = read_register(UNIV_DGCS);
#ifdef UNIVERSE_DEBUG
	printk("universe: wait %d for DMA finish %08x\n",i,dgcs);
#endif
	if ((dgcs & UNIV_BM_DGCS_ACT) == 0) break;
	/*
	  if (i<1)
	  printk("universe: after dma wait dma was not really finished!\n");
	  udelay(1);
	  */
      }
    /*
      if (i==10000)
      printk("universe: timed out waiting for DMA\n");
      */
  }
  return(dgcs);
}




static inline int 
do_dma_read(unsigned long vme_addr,
	    char *user_buffer,
	    unsigned int dctl,
	    size_t dma_irq_threshold,
	    int count)
{ /* do a DMA'd read */
  size_t bytes_to_read;
  size_t bytes_read = 0;
  char *DMA_buffer = NULL;
  int DMA_buffer_size;
  unsigned long aligned_dma_buf;
  int aligned_dma_buf_size;
#ifdef UNIVERSE_DEBUG
  printk("universe: DMA read from %08lx to %p %d bytes %08x dctl\n",
	 vme_addr, user_buffer, count, dctl);
#endif 
  /* if the maximum transfer width on the VME side is 8 or 16 bits,
     the block transfers must be disabled,
     as mentioned in the universe manual, section 2.8.1.2-3, "Transfer Size"
     and "Tranfer Data Width", around page 2-69 */
  if (((dctl & UNIV_BM_DCTL_VDW) >> UNIV_OF_DCTL_VDW) < VMEMAP_DWIDTH_32)
    dctl &= ~ UNIV_BM_DCTL_VCT;
  DMA_buffer_size = count + (vme_addr & 8);
  if (DMA_buffer_size > DMA_max_buffer_size)
    DMA_buffer_size = DMA_max_buffer_size;
#ifdef UNIVERSE_DEBUG
  printk("universe: will allocate DMA buffer of %d bytes\n",DMA_buffer_size);
#endif
  DMA_buffer = (char *)__get_dma_pages(GFP_KERNEL,__get_order(DMA_buffer_size));
#ifdef UNIVERSE_DEBUG
  printk("universe: done, addr is %p (%08lx)\n",
	 DMA_buffer, virt_to_bus(DMA_buffer));
#endif
  if (!DMA_buffer)
    {
      printk("universe: could not allocate DMA buffer");
      return(-ENOMEM);
   }

  for (bytes_to_read = count; bytes_to_read;)
    { 
      size_t bytes_read_just_now;
      int use_irq;

      aligned_dma_buf = virt_to_bus(DMA_buffer);
      aligned_dma_buf += (vme_addr % 8);
      aligned_dma_buf_size = DMA_buffer_size - (vme_addr % 8); 
      bytes_read_just_now = (bytes_to_read > aligned_dma_buf_size) ?
	aligned_dma_buf_size : bytes_to_read;
#ifdef UNIVERSE_DEBUG
      printk("universe: aligned_dma_buf: %08lx, size is %d byte\n",
	     aligned_dma_buf, aligned_dma_buf_size);
#endif

      use_irq = bytes_read_just_now > dma_irq_threshold;

#ifdef UNIVERSE_DEBUG
      printk("universe: wait for DMA semaphore\n");
#endif
      down(&DMA_available); /* wait for availability of DMA */
      /* set up local and vme address and number of bytes */
      write_register(aligned_dma_buf,UNIV_DLA);
      write_register(vme_addr,UNIV_DVA);
      write_register(bytes_read_just_now,UNIV_DTBC);
      /* set up address modifiers */
      write_register(dctl,UNIV_DCTL);
      {
	unsigned int dgcs;
	dgcs = UNIV_BM_DGCS_GO |  /* start DMA right away */
	  /* clear old error conditions */
	  UNIV_BM_DGCS_STOP | UNIV_BM_DGCS_HALT | UNIV_BM_DGCS_DONE |
	  UNIV_BM_DGCS_LERR | UNIV_BM_DGCS_VERR | UNIV_BM_DGCS_PERR |
	  /* give other busmaster a chance after each hunk of 256 bytes */
	  (UNIV_BM_DGCS_VON & (1 << UNIV_OF_DGCS_VON));
	  /* VOFF remains 0, we make round robin with the PCI slaves and
	     other bus masters */

	/* enable irq generation only if the number of bytes to be transferred
	   justifies that. For a small number it is more efficient to do
	   a polling wait! */
	if (use_irq)
	  dgcs |= UNIV_BM_DGCS_INT_DONE | 
	    UNIV_BM_DGCS_INT_LERR | UNIV_BM_DGCS_INT_VERR |
	    UNIV_BM_DGCS_INT_PERR;
	write_register(dgcs,UNIV_DGCS);
      }
      {
	unsigned int dgcs;
	dgcs = wait_for_dma_done(use_irq);
	if (dgcs & UNIV_BM_DGCS_VERR)
	  {
	    unsigned long addr;
	    int bytes_missed;
	    addr = read_register(UNIV_DVA);
	    bytes_missed = read_register(UNIV_DTBC);
#ifdef UNIVERSE_DEBUG
	    printk("universe: release DMA semaphore\n");
#endif
	    up(&DMA_available);

	    printk("universe: dma error on VME addr %08lx,"
		   " %d bytes still missing\n",
		   addr, bytes_missed);
	    bytes_read_just_now -= bytes_missed;
	    bytes_read += bytes_read_just_now;
	    memcpy_tofs(user_buffer, bus_to_virt(aligned_dma_buf), 
			bytes_read_just_now);
	    break;
	  }
	if (dgcs & UNIV_BM_DGCS_LERR)
	  {
	    unsigned long addr;
	    addr = read_register(UNIV_DLA);
	    up(&DMA_available);
	    printk("universe: dma error on PCI addr %08lx\n",addr);
	    bytes_read = -EINVAL;
	    break;
	  }
	if (dgcs & UNIV_BM_DGCS_PERR)
	  {
	    up(&DMA_available);
	    printk("universe: dma protocol error, should not happen\n");
	    bytes_read = -EINVAL;
	    break;
	  }
	/* no more access to DMA specific registers is required,
	   so mark the DMA resource as free */
	up(&DMA_available);
      }
      /* this point is reached only if the transfer completed sucessfully,
	 partial reads have their own memcpy */
      memcpy_tofs(user_buffer, bus_to_virt(aligned_dma_buf),
		  bytes_read_just_now);


      user_buffer += bytes_read_just_now;
      vme_addr += bytes_read_just_now;
      bytes_read +=bytes_read_just_now;
      bytes_to_read -= bytes_read_just_now;
    }
  free_pages((unsigned long)DMA_buffer, __get_order(DMA_buffer_size));
  DMA_buffer = NULL;
  return bytes_read;
}

static inline int 
do_dma_write(unsigned long vme_addr,
	     const char *user_buffer,
	     unsigned int dctl,
	     size_t dma_irq_threshold,
	     int count)
{ /* do a DMA'd write */
  size_t bytes_to_write;
  size_t bytes_written = 0;
  char *DMA_buffer = NULL;
  int DMA_buffer_size;
  unsigned long aligned_dma_buf;
  int aligned_dma_buf_size;
#ifdef UNIVERSE_DEBUG
  printk("universe: DMA write from %08lx to %p %d bytes %08x dctl\n",
	 vme_addr, user_buffer, count, dctl);
#endif 
  /* if the maximum transfer width on the VME side is 8 or 16 bits,
     the block transfers must be disabled,
     as mentioned in the universe manual, section 2.8.1.2-3, "Transfer Size"
     and "Tranfer Data Width", around page 2-69 */
  if (((dctl & UNIV_BM_DCTL_VDW) >> UNIV_OF_DCTL_VDW) < VMEMAP_DWIDTH_32)
    dctl &= ~ UNIV_BM_DCTL_VCT;
  dctl |= UNIV_BM_DCTL_L2V; /* set direction: VME -> PCI */
  DMA_buffer_size = count + (vme_addr & 8);
  if (DMA_buffer_size > DMA_max_buffer_size)
    DMA_buffer_size = DMA_max_buffer_size;
#ifdef UNIVERSE_DEBUG
  printk("universe: will allocate DMA buffer of %d bytes\n",DMA_buffer_size);
#endif
  DMA_buffer = (char *)__get_dma_pages(GFP_KERNEL,__get_order(DMA_buffer_size));
#ifdef UNIVERSE_DEBUG
  printk("universe: done, addr is %p (%08lx)\n",
	 DMA_buffer, virt_to_bus(DMA_buffer));
#endif
  if (!DMA_buffer)
    {
      printk("universe: could not allocate DMA buffer");
      return(-ENOMEM);
   }

  for (bytes_to_write = count; bytes_to_write;)
    { 
      int use_irq;
      size_t bytes_written_just_now;

      aligned_dma_buf = virt_to_bus(DMA_buffer);
      aligned_dma_buf += (vme_addr % 8);
      aligned_dma_buf_size = DMA_buffer_size - (vme_addr % 8); 
      bytes_written_just_now = (bytes_to_write > aligned_dma_buf_size) ?
	aligned_dma_buf_size : bytes_to_write;

      memcpy_fromfs( bus_to_virt(aligned_dma_buf), user_buffer,
		  bytes_written_just_now);

#ifdef UNIVERSE_DEBUG
      printk("universe: aligned_dma_buf: %08lx, size is %d byte\n",
	     aligned_dma_buf, aligned_dma_buf_size);
#endif

      use_irq = bytes_written_just_now > dma_irq_threshold;

#ifdef UNIVERSE_DEBUG
      printk("universe: wait for DMA semaphore\n");
#endif
      down(&DMA_available); /* wait for availability of DMA */
      /* set up local and vme address and number of bytes */
      write_register(aligned_dma_buf,UNIV_DLA);
      write_register(vme_addr,UNIV_DVA);
      write_register(bytes_written_just_now,UNIV_DTBC);
      /* set up address modifiers */
      write_register(dctl,UNIV_DCTL);
      {
	unsigned int dgcs;
	dgcs = UNIV_BM_DGCS_GO |  /* start DMA right away */
	  /* clear old error conditions */
	  UNIV_BM_DGCS_STOP | UNIV_BM_DGCS_HALT | UNIV_BM_DGCS_DONE |
	  UNIV_BM_DGCS_LERR | UNIV_BM_DGCS_VERR | UNIV_BM_DGCS_PERR |
	  /* give other busmaster a chance after each hunk of 256 bytes */
	  (UNIV_BM_DGCS_VON & (1 << UNIV_OF_DGCS_VON));
	  /* VOFF remains 0, we make round robin with the PCI slaves and
	     other bus masters */

	/* enable irq generation only if the number of bytes to be transferred
	   justifies that. For a small number it is more efficient to do
	   a polling wait! */
	if (use_irq)
	  dgcs |= UNIV_BM_DGCS_INT_DONE | 
	    UNIV_BM_DGCS_INT_LERR | UNIV_BM_DGCS_INT_VERR |
	    UNIV_BM_DGCS_INT_PERR;
	write_register(dgcs,UNIV_DGCS);
      }
      {
	unsigned int dgcs;
	dgcs = wait_for_dma_done(use_irq);
	if (dgcs & UNIV_BM_DGCS_VERR)
	  {
	    unsigned long addr;
	    int bytes_missed;
	    addr = read_register(UNIV_DVA);
	    bytes_missed = read_register(UNIV_DTBC);
#ifdef UNIVERSE_DEBUG
	    printk("universe: release DMA semaphore\n");
#endif
	    up(&DMA_available);
	    printk("universe: dma error on VME addr %08lx,"
		   " %d bytes still missing\n",
		   addr, bytes_missed);
	    bytes_written_just_now -= bytes_missed;
	    bytes_written += bytes_written_just_now;
	    break;
	  }
	if (dgcs & UNIV_BM_DGCS_LERR)
	  {
	    unsigned long addr;
	    addr = read_register(UNIV_DLA);
	    up(&DMA_available);
	    printk("universe: dma error on PCI addr %08lx\n",addr);
	    bytes_written = -EINVAL;
	    break;
	  }
	if (dgcs & UNIV_BM_DGCS_PERR)
	  {
	    up(&DMA_available);
	    printk("universe: dma protocol error, should not happen\n");
	    bytes_written = -EINVAL;
	    break;
	  }
#ifdef UNIVERSE_DEBUG
	printk("universe: release DMA semaphore\n");
#endif
	up(&DMA_available);
      }

      user_buffer += bytes_written_just_now;
      vme_addr += bytes_written_just_now;
      bytes_written +=bytes_written_just_now;
      bytes_to_write -= bytes_written_just_now;
    }
  free_pages((unsigned long)DMA_buffer, __get_order(DMA_buffer_size));
  DMA_buffer = NULL;
  return bytes_written;
}

static int 
universe_read(struct inode *inode, struct file *file, char *buf, int count)
{
  unsigned int minor = MINOR(inode->i_rdev);
  int retval;
 
  retval = verify_area(VERIFY_WRITE, buf, count);
  if (retval)
    return retval;

  switch (minor) {
  case UNIVERSE_MINOR_CTL:
    return -EINVAL;

  case UNIVERSE_MINOR_STAT:
    return read_status(buf, count);

  case UNIVERSE_MINOR_A24A16_SUPER:
  case UNIVERSE_MINOR_A24A16_USER:
  case UNIVERSE_MINOR_A32_0:
  case UNIVERSE_MINOR_A32_1:
  case UNIVERSE_MINOR_A32_2:
  case UNIVERSE_MINOR_A32_3:
    {
      struct image_desc_type *p;
      p = file->private_data;
      if (!p)
	{
	  printk("universe: missing private data in file struct\n");
	  return -EINVAL;
	}
      if (count > p->dma_threshold)
	{
	  unsigned long vme_addr;
	  vme_addr = p->vme_start + file->f_pos;
	  retval = do_dma_read(vme_addr, buf, p->dctl, 
			       p->dma_irq_threshold, count);
	}
      else
	{ /* only a few bytes, no need to do DMA */
	  switch ((p->dctl & UNIV_BM_LSI_CTL_VDW) >> UNIV_OF_LSI_CTL_VDW) {
	  case VMEMAP_DWIDTH_8:
	    retval = read_bytes((unsigned char *)buf,
				(unsigned char *)(p->kernel_addr
						  + p->mmap_offset 
						  + file->f_pos),
				count/sizeof(unsigned char));
	    
	    break;
	  case VMEMAP_DWIDTH_16:
	    retval = read_words((unsigned short *)buf,
				(unsigned short *)(p->kernel_addr
						   + p->mmap_offset 
						   + file->f_pos),
				count/sizeof(unsigned short));
	    
	    break;
	  case VMEMAP_DWIDTH_32:
	  case VMEMAP_DWIDTH_64:
	    retval = read_ints((unsigned int *)buf,
			       (unsigned int *)(p->kernel_addr
						+ p->mmap_offset 
						+ file->f_pos),
			       count/sizeof(unsigned int));
	    break;
	  }
	}
      if (retval >= 0)
	file->f_pos += retval;
      return(retval);
    }
  case UNIVERSE_MINOR_VMEIRQ:
    { /* read() on the IRQ devices means wait for an irq to occur */
      struct irq_device_type *p;
      struct wait_queue wait = { current, NULL };
      p = file->private_data;
      if (!p) return -EINVAL;
      add_wait_queue(&p->irqwait, &wait);
      current->state = TASK_INTERRUPTIBLE;
      while (p->number_of_irqs_since_last_read == 0)
	{
	  if (file->f_flags & O_NONBLOCK) {
	    retval = -EAGAIN;
	    break;
	  }
	  if (current->signal & ~current->blocked) {
	    retval = -ERESTARTSYS;
	    break;
	  }
	  schedule();
	}
      if (retval == 0) 
	{
	  unsigned long flags;
	  struct vme_irq_data_type data;
	  int statid, nof_interrupts;
	  save_flags(flags);
	  cli();
	  nof_interrupts = p->number_of_irqs_since_last_read; 
	  p->number_of_irqs_since_last_read = 0;
	  statid = p->statid; 
	  restore_flags(flags);
	  data.nof_missed_interrupts = nof_interrupts - 1;
	  if (statid & UNIV_BM_VX_STATID_ERR) /* bus error during read
						 of status/id */
	    data.vector_number = -1; /* mark as invalid */
	  else
	    data.vector_number = (statid & UNIV_BM_VX_STATID_STATID) >>
	      UNIV_OF_VX_STATID_STATID;
	  memcpy_tofs(buf, &data, sizeof(data));
	  retval = sizeof(data);
	}
      current->state = TASK_RUNNING;
      remove_wait_queue(&p->irqwait, &wait);
      return retval;
    }

  default:
    return -ENOSYS;
  }
  return 0;
}

static int 
write_bytes(const unsigned char* user_buffer,
	    unsigned char* vme_addr,
	    int count)
{
  int i;
  unsigned short pci_csr;
  for (i=0; i<count; i++)
    {
      unsigned char value;
      value = get_user(user_buffer);
      cli(); /* disable interrupts while reading value and register */ 
      write_register_word(PCI_STATUS_SIG_TARGET_ABORT, PCI_STATUS);
      writeb(value,vme_addr);
      pci_csr = read_register_word(PCI_STATUS);
      sti(); /* re-enable interrupts, the read with check is done */

      if (pci_csr & PCI_STATUS_SIG_TARGET_ABORT)
	{
	  pci_csr = PCI_STATUS_SIG_TARGET_ABORT; /* writing a 1 clears... */
	  write_register_word(pci_csr, PCI_STATUS);
	  return(-EFAULT);
	}
      else
	{
	  vme_addr++;
	  user_buffer++;
	}
    }
  return(count * sizeof(unsigned char));
} 
static int 
write_words(const unsigned short* user_buffer,
	    unsigned short* vme_addr,
	    int count)
{
  int i;
  unsigned short pci_csr;
  for (i=0; i<count; i++)
    {
      unsigned short value;
      value = get_user((unsigned short *)user_buffer);
      cli(); /* disable interrupts while reading value and register */ 
      write_register_word(PCI_STATUS_SIG_TARGET_ABORT, PCI_STATUS);
      writew(value,vme_addr);
      pci_csr = read_register_word(PCI_STATUS);
      sti(); /* re-enable interrupts, the read with check is done */

      if (pci_csr & PCI_STATUS_SIG_TARGET_ABORT)
	{
	  pci_csr = PCI_STATUS_SIG_TARGET_ABORT; /* writing a 1 clears... */
	  write_register_word(pci_csr, PCI_STATUS);
	  return(-EFAULT);
	}
      else
	{
	  vme_addr++;
	  user_buffer++;
	}
    }
  return(count * sizeof(unsigned short));
} 
static int 
write_ints(const unsigned int* user_buffer,
	    unsigned int* vme_addr,
	    int count)
{
  int i;
  unsigned short pci_csr;
  for (i=0; i<count; i++)
    {
      unsigned int value;
      value = get_user((unsigned int *)user_buffer);
      cli(); /* disable interrupts while reading value and register */ 
      write_register_word(PCI_STATUS_SIG_TARGET_ABORT, PCI_STATUS);
      writel(value,vme_addr);
      pci_csr = read_register_word(PCI_STATUS);
      sti(); /* re-enable interrupts, the read with check is done */

      if (pci_csr & PCI_STATUS_SIG_TARGET_ABORT)
	{
	  pci_csr = PCI_STATUS_SIG_TARGET_ABORT; /* writing a 1 clears... */
	  write_register_word(pci_csr, PCI_STATUS);
	  return(-EFAULT);
	}
      else
	{
	  vme_addr++;
	  user_buffer++;
	}
    }
  return(count * sizeof(unsigned int));
} 



static int 
universe_write(struct inode * inode, struct file * file,
	    const char * buf, int count)
{
  unsigned int minor = MINOR(inode->i_rdev);
  int retval;
 
  retval = verify_area(VERIFY_READ, buf, count);
  if (retval)
    return retval;
  
  switch (minor) {
  case UNIVERSE_MINOR_STAT:
  case UNIVERSE_MINOR_CTL:
    return -EINVAL;
  case UNIVERSE_MINOR_A24A16_SUPER:
  case UNIVERSE_MINOR_A24A16_USER:
  case UNIVERSE_MINOR_A32_0:
  case UNIVERSE_MINOR_A32_1:
  case UNIVERSE_MINOR_A32_2:
  case UNIVERSE_MINOR_A32_3:
    {
      struct image_desc_type *p;
      p = file->private_data;
      if (!p)
	{
	  printk("universe: missing private data in file struct\n");
	  return -EINVAL;
	}
      if (count > p->dma_threshold)
	{
	  unsigned long vme_addr;
	  vme_addr = p->vme_start + file->f_pos;
	  retval = do_dma_write(vme_addr, buf, p->dctl,
				p->dma_irq_threshold, count);
	}
      else
	{ /* only a few bytes, no need to do DMA */
	  switch ((p->dctl & UNIV_BM_LSI_CTL_VDW) >> UNIV_OF_LSI_CTL_VDW) {
	  case VMEMAP_DWIDTH_8:
	    retval = write_bytes((const unsigned char *)buf,
				 (unsigned char *)(p->kernel_addr
						   + p->mmap_offset
						   + file->f_pos),
				 count/sizeof(unsigned char));
	    break;
	  case VMEMAP_DWIDTH_16:
	    retval = write_words((const unsigned short *)buf,
				 (unsigned short *)(p->kernel_addr
						    + p->mmap_offset 
						    + file->f_pos),
				 count/sizeof(unsigned short));
	    break;
	  case VMEMAP_DWIDTH_32:
	  case VMEMAP_DWIDTH_64:
	    retval = write_ints((const unsigned int *)buf,
				(unsigned int *)(p->kernel_addr
						 + p->mmap_offset
						 + file->f_pos),
				count/sizeof(unsigned int));
	    break;
	  }
	}
      if (retval >= 0)
	file->f_pos += retval;
      return(retval);
    }
  default: break;
  }
  return -ENOSYS;
}

static int 
universe_select(struct inode *inode, struct file *file,
	     int sel_type, select_table *wait)
{
  unsigned int minor = MINOR(inode->i_rdev);
  int retval;
  switch (minor) {
  case UNIVERSE_MINOR_VMEIRQ:
    { /* read() on the IRQ devices means wait for an irq to occur */
      if (sel_type == SEL_IN)
	{
	  struct irq_device_type *p;
	  p = file->private_data;
	  if (!p) return -EINVAL;
	  if (p->number_of_irqs_since_last_read != 0)
	    return(1);
	  select_wait(&p->irqwait, wait);
	}
      return(0);
    }
  }
  return -ENOSYS;
}

static int 
universe_ioctl(struct inode *inode, struct file *file,
	       unsigned int cmd, unsigned long arg)
{
  unsigned int minor = MINOR(inode->i_rdev);
  int retval;
  switch (minor) {
  case UNIVERSE_MINOR_STAT:
    return -EINVAL; break;
  case UNIVERSE_MINOR_CTL:
    switch (cmd) {
    case VMECTL_RESET:
      {
	unsigned int misc_ctl;
	printk("universe: doing VME reset\n");
	misc_ctl = read_register( UNIV_MISC_CTL);
	misc_ctl |= UNIV_BM_MISC_CTL_SW_SRST; 
	write_register(misc_ctl, UNIV_MISC_CTL);
	printk("universe: done VME reset\n");
	retval = 0;
	break;
      }
    case VMECTL_GETVMETO:
      {
	unsigned int misc_ctl;
	int timeout;
	misc_ctl = read_register(UNIV_MISC_CTL);
	timeout = 8 << ((misc_ctl & UNIV_BM_MISC_CTL_VBTO) >> UNIV_OF_MISC_CTL_VBTO);
	if (timeout < 16) timeout = 0;
	if (timeout > 1024) timeout = -1;
	{
	put_user(timeout, (int *)arg);
	}
	retval = 0;
	break;
      }
    case VMECTL_SETVMETO:
      {
	unsigned int misc_ctl;
	int coded_timeout = 0;
	if (arg > 1024)
	  arg = 1024;
	if (arg)
	  coded_timeout = ffz(~arg)-3;
	misc_ctl = read_register(UNIV_MISC_CTL);
	misc_ctl &= ~UNIV_BM_MISC_CTL_VBTO;
	misc_ctl |= UNIV_BM_MISC_CTL_VBTO & 
	  (coded_timeout << UNIV_OF_MISC_CTL_VBTO);
	write_register(misc_ctl,UNIV_MISC_CTL);
	retval = 0;
	break;
      }
    default:
      return -EINVAL;
    }
    break;
  case UNIVERSE_MINOR_A32_0:
  case UNIVERSE_MINOR_A32_1:
  case UNIVERSE_MINOR_A32_2:
  case UNIVERSE_MINOR_A32_3:
    { /* one of the exclusive images */
      int image_no = (minor - UNIVERSE_MINOR_A32_0);
      int reg_offset = (UNIV_LSI1_CTL - UNIV_LSI0_CTL) * image_no; 
      struct image_desc_type *p;
      p = file->private_data;
      if (!p)
	{
	  printk("universe: missing private data in file struct\n");
	  return -EINVAL;
	}
      switch (cmd) {
      case VMEIMG_GETMAPPING:
	{
	  struct vme_mapping_ctrl mapctrl;
	  unsigned int ctl;
	  ctl = read_register(UNIV_LSI0_CTL + reg_offset);
	  if ((ctl & (UNIV_BM_LSI_CTL_VDW 
		     | UNIV_BM_LSI_CTL_VAS
		     | UNIV_BM_LSI_CTL_PGM
		     | UNIV_BM_LSI_CTL_SUPER
		     | UNIV_BM_LSI_CTL_VCT))
	      != p->dctl)
	    printk("universe: ctl mistmatch: reg is %08x file is %08lx\n",
		   ctl, p->dctl);
	  mapctrl.max_datawidth = (ctl & UNIV_BM_LSI_CTL_VDW) >>
	    UNIV_OF_LSI_CTL_VDW;
	  mapctrl.address_space = (ctl & UNIV_BM_LSI_CTL_VAS) >>
	    UNIV_OF_LSI_CTL_VAS;
	  mapctrl.supuseram = (ctl & UNIV_BM_LSI_CTL_SUPER) >>
	    UNIV_OF_LSI_CTL_SUPER;
	  mapctrl.prgdataam = (ctl & UNIV_BM_LSI_CTL_PGM) >>
	    UNIV_OF_LSI_CTL_PGM;
	  memcpy_tofs((char *)arg, &mapctrl, sizeof(mapctrl));
	  return 0;
	}
      case VMEIMG_SETMAPPING:
	{
	  struct vme_mapping_ctrl mapctrl;
	  unsigned int ctl;
	  memcpy_fromfs(&mapctrl, (char *)arg, sizeof(mapctrl));
	  ctl = read_register(UNIV_LSI0_CTL + reg_offset);
	  ctl &= ~(UNIV_BM_LSI_CTL_VDW | UNIV_BM_LSI_CTL_VAS
		   | UNIV_BM_LSI_CTL_SUPER | UNIV_BM_LSI_CTL_PGM);
	  ctl |= UNIV_BM_LSI_CTL_VDW & 
	    (mapctrl.max_datawidth << UNIV_OF_LSI_CTL_VDW);
	  ctl |= UNIV_BM_LSI_CTL_VAS &
	    (mapctrl.address_space << UNIV_OF_LSI_CTL_VAS);
	  ctl |= UNIV_BM_LSI_CTL_SUPER &
	    (mapctrl.supuseram << UNIV_OF_LSI_CTL_SUPER);
	  ctl |= UNIV_BM_LSI_CTL_PGM &
	    (mapctrl.prgdataam <<UNIV_OF_LSI_CTL_PGM);
	  write_register(ctl,UNIV_LSI0_CTL + reg_offset);
	  p->dctl = ctl & (UNIV_BM_LSI_CTL_VDW 
			   | UNIV_BM_LSI_CTL_VAS
			   | UNIV_BM_LSI_CTL_PGM
			   | UNIV_BM_LSI_CTL_SUPER
			   | UNIV_BM_LSI_CTL_VCT);
	  return 0;
	}
      case VMEIMG_GETVMEADDR:
	{
	  unsigned int bs, to, vme_addr;
	  bs = read_register(UNIV_LSI0_BS + reg_offset);
	  to = read_register(UNIV_LSI0_TO + reg_offset);
	  vme_addr = bs + to;
	  if (vme_addr != p->vme_start)
	    printk("universe: vme addr mismatch, reg is %08x file is %08lx\n",
		   vme_addr, p->vme_start);
	  put_user(vme_addr, (unsigned int *)arg);
	  return 0;
	}
      case VMEIMG_SETVMEADDR:
	{
	  unsigned int bs, to, vme_addr;
	  vme_addr = (unsigned int) arg;
	  /* consider granularity of the image, better zero all bits
	     that we have no control over */
	  vme_addr &= ~((image_no == 0 ? 0x00001000 : 0x00010000)-1);
	  bs = read_register(UNIV_LSI0_BS + reg_offset);
	  to = vme_addr - bs;
	  write_register(to, UNIV_LSI0_TO + reg_offset);
	  p->vme_start = vme_addr; 
	  return 0;
	}
      case VMEIMG_GETGRANULARITY:
	{
	  if (image_no == 0)
	    put_user(0x00001000,(unsigned int*)arg);
	  else
	    put_user(0x00010000,(unsigned int*)arg);
	  return 0;
	}
      }
    } /* end of A32-Image cases.
	 Notice: no break here, because the common cases, such as dma threshold
	 handling are done in the A24/A16 part */
  case UNIVERSE_MINOR_A24A16_SUPER:
  case UNIVERSE_MINOR_A24A16_USER:
    {
      struct image_desc_type *p;
      p = file->private_data;
      if (!p)
	{
	  printk("universe: missing private data in file struct\n");
	  return -EINVAL;
	}
      switch (cmd) {
      case VMEIMG_GETMAPPING:
	{
	  struct vme_mapping_ctrl mapctrl;
	  mapctrl.max_datawidth = (p->dctl & UNIV_BM_LSI_CTL_VDW) >>
	    UNIV_OF_LSI_CTL_VDW;
	  mapctrl.address_space = (p->dctl & UNIV_BM_LSI_CTL_VAS) >>
	    UNIV_OF_LSI_CTL_VAS;
	  mapctrl.supuseram = (p->dctl & UNIV_BM_LSI_CTL_SUPER) >>
	    UNIV_OF_LSI_CTL_SUPER;
	  mapctrl.prgdataam = (p->dctl & UNIV_BM_LSI_CTL_PGM) >>
	    UNIV_OF_LSI_CTL_PGM;
	  memcpy_tofs((char *)arg, &mapctrl, sizeof(mapctrl));
	  return 0;
	}
      case VMEIMG_SETMAPPING:
	{
	  struct vme_mapping_ctrl mapctrl;
	  unsigned int offset = 0;
	  unsigned int size = 0;
	  unsigned int ctl;
	  memcpy_fromfs(&mapctrl, (char *)arg, sizeof(mapctrl));
	  if (mapctrl.prgdataam != VMEMAP_PRGDATAAM_DATA)
	    return -EINVAL;
	  switch (mapctrl.max_datawidth) {
	  case VMEMAP_DWIDTH_8: return -EINVAL;
	  case VMEMAP_DWIDTH_16: offset += 0x01000000; break;
	  case VMEMAP_DWIDTH_32: offset += 0; break;
	  case VMEMAP_DWIDTH_64: return -EINVAL;
	  }
	  if (mapctrl.supuseram == VMEMAP_SUPUSERAM_SUPER)
	    offset += 0x02000000;
	  switch (mapctrl.address_space) {
	  case VMEMAP_ASPACE_A16:
	    offset += 0x00ff0000;
	    size = 0x00010000;
	    break;
	  case VMEMAP_ASPACE_A24:
	    offset += 0;
	    size = 0x00ff0000;
	    break;
	  case VMEMAP_ASPACE_A32:
	  case VMEMAP_ASPACE_CRCSR:
	  case VMEMAP_ASPACE_USR1:
	  case VMEMAP_ASPACE_USR2:
	    return -EINVAL;
	  }
	  p->map_size = size;
	  {
	    unsigned int slsi_pci_mem_start;
	    unsigned int slsi;
	    slsi = read_register( UNIV_SLSI);
	    slsi_pci_mem_start = ((slsi & UNIV_BM_SLSI_BS) >> UNIV_OF_SLSI_BS)
							   << 26;
	
	    p->phys_start = slsi_pci_mem_start +offset;
	  }
	  p->phys_end = p->phys_start + p->map_size;
	  p->kernel_addr = Slsi_kernel_addr + offset;
	  p->mmap_offset = 0;
	  p->vme_start = 0;
	  p->dctl = 0;
	  p->dctl |= UNIV_BM_LSI_CTL_VDW & 
	    (mapctrl.max_datawidth << UNIV_OF_LSI_CTL_VDW);
	  p->dctl |= UNIV_BM_LSI_CTL_VAS &
	    (mapctrl.address_space << UNIV_OF_LSI_CTL_VAS);
	  p->dctl |= UNIV_BM_LSI_CTL_SUPER &
	    (mapctrl.supuseram << UNIV_OF_LSI_CTL_SUPER);
	  p->dctl |= UNIV_BM_LSI_CTL_PGM &
	    (mapctrl.prgdataam <<UNIV_OF_LSI_CTL_PGM);
	  return 0;
	}
      case VMEIMG_GETVMEADDR:
	{
	  if (p->mmap_offset != p->vme_start)
	    printk("universe: addr mismatch\n");
	  put_user(p->vme_start, (unsigned int *)arg);
	  return 0;
	}
      case VMEIMG_SETVMEADDR:
	{
	  unsigned int vme_addr;
	  vme_addr = (unsigned int) arg;
	  
	  /* get vme_addr aligned to the page size */
	  vme_addr &= PAGE_MASK; 

	  /* the comparison with the < is no nonsense because of the
	     rules of unsigned arithmetics ! */
	  if ((p->phys_start + vme_addr >= p->phys_end)
	      || (p->phys_start + vme_addr < p->phys_start))
	    {
	      printk("universe: a24/16 setvme %x ps %x pe %x",
		     vme_addr, p->phys_start, p->phys_start);
	      return -EINVAL;
	    }
	  p->mmap_offset = vme_addr;
	  p->vme_start = vme_addr; 
	  return 0;
	}
      case VMEIMG_GETGRANULARITY:
	{
	  unsigned long granularity = PAGE_SIZE;
	  put_user(granularity,(unsigned int*)arg);
	  return 0;
	}
      /* now the cases specific to the A24/16 images are handled */

      /* because of there being no break at the end of the A32 cases
	 they too may reach here, wich is just fine because the cases
	 handled here are the same for all image types */
      case VMEIMG_GETDMATRESHOLD:
	{
	  put_user(p->dma_threshold,(unsigned int*)arg);
	  retval = 0;
	  break;
	}
      case VMEIMG_SETDMATRESHOLD:
	{
	  p->dma_threshold = (unsigned long)arg;
	  retval = 0;
	  break;
	}
      default:
	return -EINVAL;
      }
      break;
    }
  case UNIVERSE_MINOR_VMEIRQ:
    {
      struct irq_device_type *p;
      p = file->private_data;
      switch (cmd) {
      case VMEIRQ_GETLINK:
	{
	  struct vme_irq_mapping_type md = {0,0};
	  if (p)
	    {
	      md.level = p->level;
	      md.vector = p->vector;
	    }
	  memcpy_tofs((char *)arg, &md, sizeof(md));
	  return(0);
	}
      case VMEIRQ_SETLINK:
	{
	  struct vme_irq_mapping_type md;
	  memcpy_fromfs(&md, (char*)arg, sizeof(md));
	  if (p)
	    p->busy = 0;
	  file->private_data = NULL;
	  if (md.level < 0 || md.level > 7 || md.vector < 0 || md.vector > 255)
	    return -EINVAL; /* illegal level or vector */
	  if (irq_device[md.level -1][md.vector].busy)
	    return -EBUSY;  /* this combination is already routed */
	  irq_device[md.level -1][md.vector].busy = 1;
	  file->private_data = &irq_device[md.level -1][md.vector];
	  return(0);
	}
      default:
	return -EINVAL;
      }
    }
  default:
    return -EINVAL;
  }
  return retval;
}

static int 
universe_mmap(struct inode *inode, struct file *file,
			 struct vm_area_struct *vma)
{
  unsigned int minor = MINOR(inode->i_rdev);
  int retval;
  
  switch (minor) {
  case UNIVERSE_MINOR_CTL:
  case UNIVERSE_MINOR_STAT:
    return -EINVAL;
  case UNIVERSE_MINOR_A24A16_SUPER:
  case UNIVERSE_MINOR_A24A16_USER:
  case UNIVERSE_MINOR_A32_0:
  case UNIVERSE_MINOR_A32_1:
  case UNIVERSE_MINOR_A32_2:
  case UNIVERSE_MINOR_A32_3:
    {
      struct image_desc_type *p;
      p = file->private_data;
#ifdef UNIVERSE_DEBUG
      printk("universe_mmap: start at 0x%08lx  offset 0x%08lx   end 0x%08lx, slsi 0x%08x\n",
	     vma->vm_start, vma->vm_offset, vma->vm_end, p->phys_start);
#endif
      if (vma->vm_end - vma->vm_start + vma->vm_offset + p->mmap_offset> 
	  p->phys_end - p->phys_start)
	{ /* an atempt to map beyond the end of the image */
	  printk("universe_mmap: INVALID "
		 "start at 0x%08lx offset 0x%08lx end 0x%08lx,"
		 "pstart 0x%08lx, pend 0x%08lx, mmapo 0x%08lx\n",
		 vma->vm_start, vma->vm_offset, vma->vm_end,
		 p->phys_start,p->phys_end,p->mmap_offset);
	  return -EINVAL;
	}
      retval = remap_page_range(vma->vm_start,
				vma->vm_offset 
				+ p->phys_start + p->mmap_offset,
				vma->vm_end - vma->vm_start,
				vma->vm_page_prot);
      if (retval) 
	{
	  printk("universe_mmap: remap_page_range failed");
	  return (-EAGAIN);
	}
      vma->vm_inode = inode;
      inode->i_count++;
      return 0;
    }
  default: return -ENODEV;
  }
  return 0;
}


static int 
universe_open(struct inode * inode, struct file * file)
{
  unsigned int minor = MINOR(inode->i_rdev);

  switch (minor) {
  case UNIVERSE_MINOR_STAT:
    if (Status_buffer)
      return -EBUSY;
    Status_buffer = kmalloc(UNIVERSE_STATUS_BUFFER_SIZE, GFP_KERNEL);
    if (!Status_buffer)
      return -EIO;
    Status_length = 0;
    Status_pointer = 0;
    init_status();
    break;
  case UNIVERSE_MINOR_CTL:
    opens_of_ctl_device++;
    break;
  case UNIVERSE_MINOR_A24A16_USER:
  case UNIVERSE_MINOR_A24A16_SUPER:
    {
      unsigned int slsi_pci_mem_start = 0;
      file->private_data = kmalloc(sizeof(struct image_desc_type),
				   GFP_KERNEL);
      if (!(file->private_data))
	return -ENOMEM;
      if (!opens_of_slsi_image)
	{ /* now initialize the special PCI slave image used for A16/A24 
	     VME-Address range */
	  unsigned int slsi;
	  /* enable special image */
	  slsi = UNIV_BM_SLSI_EN;
	  /* set first 16MB area to 32 Bit, second to 16 Bit,
	     third to 32 and fourth to 16 again. */
	  slsi |= (0x05 << UNIV_OF_SLSI_VDW) & UNIV_BM_SLSI_VDW;
	  /* set first 2 16MB areas to user, 3rd and 4.th to SUPERvisor AMcode */
	  slsi |= (0x0C << UNIV_OF_SLSI_SUPER) & UNIV_BM_SLSI_SUPER;
	  slsi_pci_mem_start = 0x10000000;
	  slsi |= ((slsi_pci_mem_start >> 26) << UNIV_OF_SLSI_BS) 
	    & UNIV_BM_SLSI_BS;
	  write_register(slsi, UNIV_SLSI);
	  if (!Slsi_kernel_addr)
	    Slsi_kernel_addr = ioremap(slsi_pci_mem_start,64*1024*1024);
	if (!Slsi_kernel_addr)
	  {
	    printk("universe: ioremap of slsi image failed\n");
	    return -ENOMEM;
	  }
	}
      opens_of_slsi_image++;
      {
	struct image_desc_type *p;
	size_t offset = 0;
	p = file->private_data;
	p->dma_threshold = _DMA_threshold;
	p->dma_irq_threshold = _DMA_irq_threshold;
	switch (minor) {
	case UNIVERSE_MINOR_A24A16_SUPER:
	  p->kernel_addr = Slsi_kernel_addr + SLSI_A24D32_S_OFFSET;
	  offset = SLSI_A24D32_S_OFFSET;
	  p->map_size = 0x00ff0000;
	  p->dctl = (UNIV_BM_DCTL_VDW & (VMEMAP_DWIDTH_32 << UNIV_OF_DCTL_VDW))
	    | (UNIV_BM_DCTL_VAS & (VMEMAP_ASPACE_A24 << UNIV_OF_DCTL_VAS))
	    | (UNIV_BM_DCTL_PGM & (VMEMAP_PRGDATAAM_DATA << UNIV_OF_DCTL_PGM))
	    | (UNIV_BM_DCTL_SUPER &(VMEMAP_SUPUSERAM_SUPER<<UNIV_OF_DCTL_SUPER));
	  break;
	case UNIVERSE_MINOR_A24A16_USER:
	  p->kernel_addr = Slsi_kernel_addr + SLSI_A24D32_OFFSET;
	  offset = SLSI_A24D32_OFFSET;
	  p->map_size = 0x00ff0000;
	  p->dctl = (UNIV_BM_DCTL_VDW & (VMEMAP_DWIDTH_32 << UNIV_OF_DCTL_VDW))
	    | (UNIV_BM_DCTL_VAS & (VMEMAP_ASPACE_A24 << UNIV_OF_DCTL_VAS))
	    | (UNIV_BM_DCTL_PGM & (VMEMAP_PRGDATAAM_DATA << UNIV_OF_DCTL_PGM))
	    | (UNIV_BM_DCTL_SUPER & (VMEMAP_SUPUSERAM_USER<<UNIV_OF_DCTL_SUPER));
	  break;
	}
	p->mmap_offset = 0;
	p->phys_start = slsi_pci_mem_start +offset;
	p->phys_end = p->phys_start + p->map_size;
	p->vme_start = 0;
      }
      break;
    }
  case UNIVERSE_MINOR_A32_0:
  case UNIVERSE_MINOR_A32_1:
  case UNIVERSE_MINOR_A32_2:
  case UNIVERSE_MINOR_A32_3:
    { /* one of the exclusive images */
      int image_no = (minor - UNIVERSE_MINOR_A32_0);
      int reg_offset = (UNIV_LSI1_CTL - UNIV_LSI0_CTL) * image_no; 
      {
	unsigned int ctl;
	unsigned int pci_start_addr;
	ctl = read_register(UNIV_LSI0_CTL + reg_offset);
	if (ctl & UNIV_BM_LSI_CTL_EN) /* device enabled: busy busy! */
	  return -EBUSY;
	/* not busy, so programm sensible default values */

	pci_start_addr = UNIVERSE_A32_BASE_START+UNIVERSE_A32_SPACING*image_no;
	write_register(pci_start_addr,UNIV_LSI0_BS + reg_offset);
	write_register(pci_start_addr+UNIVERSE_A32_SPACING,
		       UNIV_LSI0_BD + reg_offset);
	write_register(0u - pci_start_addr,UNIV_LSI0_TO + reg_offset);
	a32_img[image_no].vme_start = 0;
	a32_img[image_no].phys_start = pci_start_addr;
	a32_img[image_no].phys_end = pci_start_addr + UNIVERSE_A32_SPACING;
	a32_img[image_no].mmap_offset = 0;

	ctl = UNIV_BM_LSI_CTL_EN; /* enable the image */
	/* set max VME data width to 32Bit */
	ctl |= (0x02 << UNIV_OF_LSI_CTL_VDW) & UNIV_BM_LSI_CTL_VDW;
	/* set VME Adress space to A32 */
	ctl |= (0x02 << UNIV_OF_LSI_CTL_VAS) & UNIV_BM_LSI_CTL_VAS;
	write_register(ctl,UNIV_LSI0_CTL + reg_offset);

	/* most of the dctl settings are just copies of the ctl bits */
	a32_img[image_no].dctl = ctl & (UNIV_BM_LSI_CTL_VDW 
					| UNIV_BM_LSI_CTL_VAS
					| UNIV_BM_LSI_CTL_PGM
					| UNIV_BM_LSI_CTL_SUPER
					| UNIV_BM_LSI_CTL_VCT);

	if (a32_img[image_no].kernel_addr) /* should not happen! */
	  {
	    printk("universe: open: Oh! image %d has an iomap\n",image_no);
	    iounmap(a32_img[image_no].kernel_addr);
	  }
	a32_img[image_no].kernel_addr = ioremap(pci_start_addr,
						UNIVERSE_A32_DEFAULT_MAP_SIZE);
	a32_img[image_no].map_size = UNIVERSE_A32_DEFAULT_MAP_SIZE;
	a32_img[image_no].dma_irq_threshold = _DMA_irq_threshold;
	a32_img[image_no].dma_threshold = _DMA_threshold;
	file->private_data = &a32_img[image_no];
      }
      break;
    }
  case UNIVERSE_MINOR_VMEIRQ:
    {
      file->private_data = NULL;
      break;
    }
  default:
    return -ENODEV;
  }
  MOD_INC_USE_COUNT;
  return 0;
}

static void 
universe_release(struct inode * inode, struct file * file)
{
  unsigned int minor = MINOR(inode->i_rdev);

  switch (minor) {
  case UNIVERSE_MINOR_STAT:
    kfree_s(Status_buffer,UNIVERSE_STATUS_BUFFER_SIZE);
    Status_buffer = NULL;
    break;
  case UNIVERSE_MINOR_CTL:
    opens_of_ctl_device--;
    break;
  case UNIVERSE_MINOR_A24A16_SUPER:
  case UNIVERSE_MINOR_A24A16_USER:
    {
      kfree(file->private_data);
      file->private_data = NULL;

      opens_of_slsi_image--;
      if (!opens_of_slsi_image)
	{
	  iounmap(Slsi_kernel_addr);
	  Slsi_kernel_addr = NULL;
	}
      break;
    }
  case UNIVERSE_MINOR_A32_0:
  case UNIVERSE_MINOR_A32_1:
  case UNIVERSE_MINOR_A32_2:
  case UNIVERSE_MINOR_A32_3:
    { /* one of the exclusive images */
      int image_no = (minor - UNIVERSE_MINOR_A32_0);
      int reg_offset = (UNIV_LSI1_CTL - UNIV_LSI0_CTL) * image_no; 
      {
	unsigned int ctl;
	ctl = read_register(UNIV_LSI0_CTL + reg_offset);
	if (!(ctl & UNIV_BM_LSI_CTL_EN)) /* device not enabled: strange! */
	  {
	    printk("universe: release: PCI slave image %d not busy !?\n",
		   image_no);
	    return;
	  }
	ctl = 0;
	write_register(ctl, UNIV_LSI0_CTL + reg_offset);
	if (a32_img[image_no].kernel_addr == NULL)
	  printk("universe: close: oh! image %d has no iomapping\n", image_no);
	iounmap(a32_img[image_no].kernel_addr);
	a32_img[image_no].kernel_addr = NULL;
      }
      break;
    }
  case UNIVERSE_MINOR_VMEIRQ:
    {
      struct irq_device_type *p;
      p = file->private_data;
      if (!p) break;
      if (!p->busy)
	  printk("universe: Oh shit! trying to close non open IRQ device\n");
      p->busy=0; /* mark as free */
      break;
    }
  default:
    printk("universe: unknown device (minor %d) closed\n", minor);
    return;
  }
  MOD_DEC_USE_COUNT;
}


/*
  ------------------------------------------------------------------------
  Interrupt handler
  */
static inline void
handle_vme_irq(int level)
{
  int statid;
  int vector;
  Interrupt_counters.virq[level-1]++;
  statid = read_register(UNIV_V1_STATID + 
			 (UNIV_V2_STATID - UNIV_V1_STATID) * (level-1));
  if (statid & UNIV_BM_VX_STATID_ERR) /* bus error during IACK cycle */
    {
    }
  else
    {
      vector = (statid & UNIV_BM_VX_STATID_STATID) >>
	UNIV_OF_VX_STATID_STATID;
      irq_device[level-1][vector].number_of_irqs_since_last_read++;
      wake_up_interruptible(&irq_device[level-1][vector].irqwait);
    }
}

static void
universe_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
  unsigned int lint_stat;
  unsigned int lint_estat;
  unsigned int lint_en;
  int multiplicity=0;
  lint_stat = read_register(UNIV_LINT_STAT); /* read the interrupt bits */
  write_register(lint_stat,UNIV_LINT_STAT);  /* and clear again! */
  lint_en = read_register(UNIV_LINT_EN);     /* read enabled irq bits */
  lint_estat = lint_stat & lint_en;          /* and look only at enabled irqs*/
  if (lint_estat & UNIV_BM_LINT_ACFAIL)
    {
      Interrupt_counters.acfail++;
      multiplicity++;
    }
  if (lint_estat & UNIV_BM_LINT_SYSFAIL) 
    {
      Interrupt_counters.sysfail++;
      multiplicity++;
    }
  if (lint_estat & UNIV_BM_LINT_SW_INT)
    {
      Interrupt_counters.sw_int++;
      multiplicity++;
    }
  if (lint_estat & UNIV_BM_LINT_SW_IACK)
    {
      Interrupt_counters.sw_iack++;
      multiplicity++;
    }
  if (lint_estat & UNIV_BM_LINT_VERR)
    {
      Interrupt_counters.verr++;
      multiplicity++;
    }
  if (lint_estat & UNIV_BM_LINT_LERR)
    {
      Interrupt_counters.lerr++;
      multiplicity++;
    }
  if (lint_estat & UNIV_BM_LINT_DMA)
    {
      Interrupt_counters.dma++;
      multiplicity++;
      wake_up_interruptible(&DMA_wait);
    }
  if (lint_estat & UNIV_BM_LINT_VIRQ7)
    {
      handle_vme_irq(7);
      multiplicity++;
    }
  if (lint_estat & UNIV_BM_LINT_VIRQ6)
    {
      handle_vme_irq(6);
      multiplicity++;
    }
  if (lint_estat & UNIV_BM_LINT_VIRQ5)
    {
      handle_vme_irq(5);
      multiplicity++;
    }
  if (lint_estat & UNIV_BM_LINT_VIRQ4)
    {
      handle_vme_irq(4);
      multiplicity++;
    }
  if (lint_estat & UNIV_BM_LINT_VIRQ3)
    {
      handle_vme_irq(3);
      multiplicity++;
    }
  if (lint_estat & UNIV_BM_LINT_VIRQ2)
    {
      handle_vme_irq(2);
      multiplicity++;
    }
  if (lint_estat & UNIV_BM_LINT_VIRQ1)
    {
      handle_vme_irq(1);
      multiplicity++;
    }
  if (lint_estat & UNIV_BM_LINT_VOWN)
    {
      Interrupt_counters.vown++;
      multiplicity++;
    }

}


/*
  ------------------------------------------------------------------------
  Status handling functions, from sound driver.
 */
static int
put_status (char *s)
{
  int             l = strlen (s);
 
  if (Status_length + l >= UNIVERSE_STATUS_BUFFER_SIZE)
    return 0;
  memcpy(&Status_buffer[Status_length], s, l);
  Status_length += l;
  return 1;
}

static int
put_status_int (unsigned int val, int radix)
{
  int             l, v;
  static char     hx[] = "0123456789ABCDEF";
  char            buf[11];
  if (!val)
    return put_status ("0");
  l = 0;
  buf[10] = 0;
  while (val)
    {
      v = val % radix;
      val = val / radix;
      buf[9 - l] = hx[v];
      l++;
    }
  if (Status_length + l >= UNIVERSE_STATUS_BUFFER_SIZE)
    return 0;
  memcpy (&Status_buffer[Status_length], &buf[10 - l], l);
  Status_length += l;
  return 1;
}

static void
init_status(void)
{
  put_status("Universe PCI/VME Bridge Status\nDriver $Id: universe.c,v 1.43 1997/06/09 10:46:00 hannappe Exp $\nCompiled "
	     __DATE__ " " __TIME__ "\n");
  put_status("UCSR_base_addr is "); 
  put_status_int(read_register( PCI_BASE_ADDRESS_0) 
		 & PCI_BASE_ADDRESS_MEM_MASK,16);
  put_status(" Interrupt line is "); 
  put_status_int(read_register(UNIV_PCI_MISC1) & 0x000000FF, 10);
  put_status(", Latency timer ");
  put_status_int((read_register(UNIV_PCI_MISC0) & UNIV_BM_PCI_MISC0_LTIMER)
		 >> UNIV_OF_PCI_MISC0_LTIMER,10);
  put_status(" clocks\n");
  {
    unsigned long misc_ctl;
    misc_ctl = read_register( UNIV_MISC_CTL);
    put_status("Misc_Ctl is "); put_status_int(misc_ctl,16);
    if (misc_ctl & UNIV_BM_MISC_CTL_SYSCON)
      put_status(", VME system controller");
    if (misc_ctl & UNIV_BM_MISC_CTL_RESCIND)
      put_status(", DTACK Rescind enabled");
    put_status("\n");
    put_status("VME time out ");
    switch ((misc_ctl & UNIV_BM_MISC_CTL_VBTO) >> UNIV_OF_MISC_CTL_VBTO) {
    case 0x0: put_status("disabled"); break;
    case 0x1: put_status("16 us"); break;
    case 0x2: put_status("32 us"); break;
    case 0x3: put_status("64 us"); break;
    case 0x4: put_status("128 us"); break;
    case 0x5: put_status("256 us"); break;
    case 0x6: put_status("512 us"); break;
    case 0x7: put_status("1024 us"); break;
    default: put_status("Reserved Value, undefined");
    }
    put_status(", VME arbitration time out ");
    switch ((misc_ctl & UNIV_BM_MISC_CTL_VARBTO) >> UNIV_OF_MISC_CTL_VARBTO) {
    case 0x0: put_status("disabled"); break;
    case 0x1: put_status("16 us"); break;
    case 0x2: put_status("256 us"); break;
    default: put_status("Reserved Value, undefined");
    }
    if (misc_ctl & UNIV_BM_MISC_CTL_VARB)
      put_status(", Priority");
    else
      put_status(", Round Robin");
    put_status(" arbitration");
    put_status("\n");
  }
  {
    unsigned int lmisc;
    unsigned int crt,cwt;
    lmisc = read_register(UNIV_LMISC);  
    crt = (lmisc & UNIV_BM_LMISC_CRT) >> UNIV_OF_LMISC_CRT;
    cwt = (lmisc & UNIV_BM_LMISC_CWT) >> UNIV_OF_LMISC_CWT;
    put_status("Coupled request timer ");
    switch (crt) {
    case 0x0: put_status("disabled"); break;
    case 0x1: put_status(" 128"); break;
    case 0x2: put_status(" 256"); break;
    case 0x3: put_status(" 512"); break;
    case 0x4: put_status("1024"); break;
    case 0x5: put_status("2048"); break;
    case 0x6: put_status("4096"); break;
    default: put_status(" Reserved");
    }
    put_status(" us, Coupled Window timer ");
    switch (crt) {
    case 0x0: put_status("disabled"); break;
    case 0x1: put_status(" 16"); break;
    case 0x2: put_status(" 32"); break;
    case 0x3: put_status(" 64"); break;
    case 0x4: put_status("128"); break;
    case 0x5: put_status("256"); break;
    case 0x6: put_status("512"); break;
    default: put_status(" Reserved");
    }
    put_status(" PCI clocks\n");
  }
  {
    unsigned int mast_ctl;
    mast_ctl = read_register( UNIV_MAST_CTL);
    {
      int retries;
      retries = ((mast_ctl & UNIV_BM_MAST_CTL_MAXRTRY) 
	>> UNIV_OF_MAST_CTL_MAXRTRY) * 64;
      put_status("Max PCI Master Retries: ");
      if (retries)
	put_status_int(retries,10);
      else
	put_status("forever");
    }

    switch ((mast_ctl & UNIV_BM_MAST_CTL_PWON) >> UNIV_OF_MAST_CTL_PWON) {
    case 0x0: put_status(", 128 Bytes"); break;
    case 0x1: put_status(", 256 Bytes"); break;
    case 0x2: put_status(", 512 Bytes"); break;
    case 0x3: put_status(", 1024 Bytes"); break;
    case 0x4: put_status(", 2028 Bytes"); break;
    case 0x5: put_status(", 4096 Bytes"); break;
    default: put_status(", undefined");
    }
    put_status(" Posted Write transfer count\n");

    switch ((mast_ctl & UNIV_BM_MAST_CTL_VRL) >> UNIV_OF_MAST_CTL_VRL) {
    case 0x0: put_status("VME Brq 0");
    case 0x1: put_status("VME Brq 1");
    case 0x2: put_status("VME Brq 2");
    case 0x3: put_status("VME Brq 3");
    }
    if (mast_ctl & UNIV_BM_MAST_CTL_VRM)
      put_status(", Fair Request mode");
    else
      put_status(", Demand Request mode");
    if (mast_ctl & UNIV_BM_MAST_CTL_VREL)
      put_status(", Release when done");
    else
      put_status(", Release on Request");
    if (mast_ctl & UNIV_BM_MAST_CTL_VOWN_ACK)
      put_status(", Owning VMEBus\n");
    else
      put_status(", not owning VMEBus\n");
  }
  {
    unsigned int misc_stat;
    misc_stat = read_register( UNIV_MISC_STAT);
    put_status("Universe BBSY");
    if (misc_stat & UNIV_BM_MISC_STAT_MYBBSY)
      put_status(" asserted");
    else
      put_status(" negated");
    put_status(", Transmit FIFO");
    if (misc_stat & UNIV_BM_MISC_STAT_TXFE)
      put_status(" empty");
    else
      put_status(" not empty");
    put_status(", Receive FIFO");
    if (misc_stat & UNIV_BM_MISC_STAT_RXFE)
      put_status(" empty\n");
    else
      put_status(" not empty\n");
  }
  {
    unsigned long vrai_ctl;
    vrai_ctl = read_register( UNIV_VRAI_CTL);
    if (vrai_ctl & UNIV_BM_VRAI_CTL_EN)
      {
	unsigned int vrai_bs;
	vrai_bs = read_register( UNIV_VRAI_BS);
	put_status("VME Register image enabled at VME-Address 0x");
	put_status_int(vrai_bs,16);
	put_status("\n");
      }
    else
      put_status("VME Register image disabled\n");
  }
  {
    unsigned int lint_en;
    unsigned int lint_stat;
    lint_en = read_register(UNIV_LINT_EN);
    lint_stat = read_register(UNIV_LINT_STAT);
#define REPORT_IRQ(name,field)     put_status(#name "\tinterupt "); \
    put_status((lint_en & UNIV_BM_LINT_##name) ? "enabled" : " masked"); \
    put_status(", triggered "); \
    put_status_int(Interrupt_counters.field,10); \
    put_status(" times"); \
    put_status((lint_stat & UNIV_BM_LINT_##name) ? ", irq now active\n" : "\n");
    
    REPORT_IRQ(ACFAIL,acfail);
    REPORT_IRQ(SYSFAIL,sysfail);
    REPORT_IRQ(SW_INT,sw_int);
    REPORT_IRQ(SW_IACK,sw_iack);
    REPORT_IRQ(VERR,verr);
    REPORT_IRQ(LERR,lerr);
    REPORT_IRQ(DMA,dma);
    REPORT_IRQ(VIRQ7,virq[7-1]);
    REPORT_IRQ(VIRQ6,virq[6-1]);
    REPORT_IRQ(VIRQ5,virq[5-1]);
    REPORT_IRQ(VIRQ4,virq[4-1]);
    REPORT_IRQ(VIRQ3,virq[3-1]);
    REPORT_IRQ(VIRQ2,virq[2-1]);
    REPORT_IRQ(VIRQ1,virq[1-1]);
    REPORT_IRQ(VOWN,vown);
#undef REPORT_IRQ
  }
  {
      unsigned int slsi;

      slsi = read_register( UNIV_SLSI);
      if (slsi & UNIV_BM_SLSI_EN)
	{
	  unsigned int slsi_pci_mem_start;
	  slsi_pci_mem_start = ((slsi & UNIV_BM_SLSI_BS) >> UNIV_OF_SLSI_BS)
							 << 26;
	  put_status("Special PCI Slave image enabled, SLSI is 0x");
	  put_status_int(slsi,16);
	  if (slsi & UNIV_BM_SLSI_PWEN)
	    put_status(", posted write enabled");
	  put_status("\nPCI base adress is 0x");
	  put_status_int(slsi_pci_mem_start,16);
	  put_status(", mapped to kernel address 0x");
	  put_status_int((int)Slsi_kernel_addr,16);
	  put_status("\n");
	  {
	    int region, region_bit;
	    for (region = 0, region_bit = 0x1; region < 4; 
		 region++, (region_bit <<= 1))
	      {
		int data;
		put_status("\tRegion "); put_status_int(region,10);
		data = (slsi & UNIV_BM_SLSI_VDW) >> UNIV_OF_SLSI_VDW;
		if (data & region_bit)
		  put_status(": 32-Bit VME ");
		else
		  put_status(": 16-Bit VME ");

		data = (slsi & UNIV_BM_SLSI_PGM) >> UNIV_OF_SLSI_PGM;
		if (data & region_bit)
		  put_status("Progam Code ");
		else
		  put_status("just Data   ");
		
		data = (slsi & UNIV_BM_SLSI_SUPER) >> UNIV_OF_SLSI_SUPER;
		if (data & region_bit)
		  put_status("Privileged Acces\n");
		else
		  put_status("User Access\n");
	      }
	  }
	}
      else
	put_status("Special PCI Slave image disabled\n");
  }

  {
    int i;
    for (i=0; i<4; i++)
      {
	unsigned int ctl, bs, bd, to, vstart, vend;
	ctl = read_register( UNIV_LSI0_CTL + 
		    i * (UNIV_LSI1_CTL-UNIV_LSI0_CTL));
	bs = read_register( UNIV_LSI0_BS + 
		    i * (UNIV_LSI1_CTL-UNIV_LSI0_CTL));
	bd = read_register( UNIV_LSI0_BD + 
		    i * (UNIV_LSI1_CTL-UNIV_LSI0_CTL));
	to = read_register( UNIV_LSI0_TO + 
		    i * (UNIV_LSI1_CTL-UNIV_LSI0_CTL));
	vstart = bs+to;
	vend = bs+to;
	put_status("PCI Slave image ");
	put_status_int(i,10);
	if (ctl & UNIV_BM_LSI_CTL_EN)
	  {
	    put_status(" enabled");
	    if (ctl & UNIV_BM_LSI_CTL_PWEN)
	      put_status(", posted write enabled\n");
	    else
	      put_status("\n");
	    put_status("PCI Addresses from 0x");
	    put_status_int(bs,16);
	    put_status(" to 0x");
	    put_status_int(bd,16);
	    put_status("\nVME Addresses from 0x");
	    put_status_int(vstart,16);
	    put_status(" to 0x");
	    put_status_int(vend,16);
	    put_status("\n");
	  }
	else
	  put_status(" disabled\n");
      }
  }

  {
    int i;
    for (i=0; i<4; i++)
      {
	unsigned int ctl, bs, bd, to, vstart, vend;
	ctl = read_register( UNIV_VSI0_CTL + 
		    i * (UNIV_VSI1_CTL-UNIV_VSI0_CTL));
	bs = read_register( UNIV_VSI0_BS + 
		    i * (UNIV_VSI1_CTL-UNIV_VSI0_CTL));
	bd = read_register( UNIV_VSI0_BD + 
		    i * (UNIV_VSI1_CTL-UNIV_VSI0_CTL));
	to = read_register( UNIV_VSI0_TO + 
		    i * (UNIV_VSI1_CTL-UNIV_VSI0_CTL));
	vstart = bs+to;
	vend = bs+to;
	put_status("VME Slave image ");
	put_status_int(i,10);
	if (ctl & UNIV_BM_LSI_CTL_EN)
	  {
	    put_status(" enabled");
	    if (ctl & UNIV_BM_LSI_CTL_PWEN)
	      put_status(", posted write enabled\n");
	    else
	      put_status("\n");
	    put_status("VME Addresses from 0x");
	    put_status_int(bs,16);
	    put_status(" to 0x");
	    put_status_int(bd,16);
	    put_status("\nPCI Addresses from 0x");
	    put_status_int(vstart,16);
	    put_status(" to 0x");
	    put_status_int(vend,16);
	    put_status("\n");
	  }
	else
	  put_status(" disabled\n");
      }
  }

}

static int
read_status (char *buf, int count)
{
  /*
   * Return at most 'count' bytes from the status_buf.
   */
  int             l, c;

  l = count;
  c = Status_length - Status_pointer;

  if (l > c)
    l = c;
  if (l <= 0)
    return 0;

  memcpy_tofs (&(buf)[0], &Status_buffer[Status_pointer], l);
  Status_pointer += l;

  return l;
}

/*
  -------------------------------------------------------------------
 */

/*
  the initialisation routine
 */
int 
init_module(void)
{
  printk("universe: $Id: universe.c,v 1.43 1997/06/09 10:46:00 hannappe Exp $\n");
  if (register_chrdev(UNIVERSE_MAJOR,"universe",&universe_fops)) 
    {
      printk("universe: unable to get major %d\n", UNIVERSE_MAJOR);
      return -EIO;
    }

  if (!pcibios_present())
    {
      printk("universe: no pcibios present, giving up\n");
      unregister_chrdev(UNIVERSE_MAJOR,"universe");
      return -ENOSYS;
    }
  {
    int found = 0;
    int pci_index;
    for (pci_index = 0; pci_index<8; pci_index++)
      {
	unsigned char pci_bus;
	unsigned char pci_device_fn;
	unsigned int pci_ioaddr;
	
	if (pcibios_find_device(PCI_VENDOR_ID_TUNDRA,
				PCI_DEVICE_ID_TUNDRA_UNIVERSE,
				pci_index,
				&pci_bus, &pci_device_fn)) 
	  continue;
	found = 1;
	printk("universe: found at bus %x device %x index %x\n",
	       pci_bus, pci_device_fn, pci_index);
	pcibios_read_config_dword(pci_bus, pci_device_fn,
				  PCI_BASE_ADDRESS_0, &pci_ioaddr);
	if ((pci_ioaddr & PCI_BASE_ADDRESS_SPACE) == PCI_BASE_ADDRESS_SPACE_IO)
	  {
	    printk(KERN_INFO "universe: the device uses io addresses\n");
#ifdef UNIVERSE_REGISTERS_ARE_MEMORY_MAPPED
	    printk(KERN_ERR "universe: But this driver is compiled for memory
 mapped access\n");
	    unregister_chrdev(UNIVERSE_MAJOR,"universe");
	    return(-EIO);
#else
	    UCSR_base_addr = pci_ioaddr & PCI_BASE_ADDRESS_IO_MASK;
#endif	    
	  }
	if ((pci_ioaddr & PCI_BASE_ADDRESS_SPACE) == PCI_BASE_ADDRESS_SPACE_MEMORY)
	  {
	    printk("universe: the device uses memory address: %x\n",
		   pci_ioaddr & PCI_BASE_ADDRESS_MEM_MASK);
#ifndef UNIVERSE_REGISTERS_ARE_MEMORY_MAPPED
	    printk(KERN_ERR "universe: But this driver is compiled for IO mapped access\n");
	    unregister_chrdev(UNIVERSE_MAJOR,"universe");
	    return(-EIO);
#else
	    /*
	    printk("universe: Setting Base adress as recomended by ELTEC\n");
	    pci_ioaddr = (0x08000000 & PCI_BASE_ADDRESS_MEM_MASK)
	      | (pci_ioaddr & (~PCI_BASE_ADDRESS_MEM_MASK));
	    pcibios_write_config_dword(pci_bus, pci_device_fn,
				  PCI_BASE_ADDRESS_0, pci_ioaddr);
	    UCSR_base_addr = ioremap(0x08000000,4*1024);
	    */
	    UCSR_base_addr = ioremap(pci_ioaddr & PCI_BASE_ADDRESS_MEM_MASK,
				     4*1024);
#endif	    	    
	  }

	printk("universe: The device id is %08x, the interrupt is %d\n",
	       read_register(UNIV_PCI_ID),
	       read_register(UNIV_PCI_MISC1) & 0x000000FF);
	{
	  unsigned long misc_ctl;
	  misc_ctl = read_register( UNIV_MISC_CTL);
	  printk("universe: Misc_Ctl is %08lx\n",misc_ctl);
	  if (misc_ctl & UNIV_BM_MISC_CTL_SYSCON)
	    printk("universe: is VME system controller\n");
	}
	printk("universe: disabling all images\n");
	write_register(0, UNIV_SLSI);
	write_register(0, UNIV_LSI0_CTL);
	write_register(0, UNIV_LSI1_CTL);
	write_register(0, UNIV_LSI2_CTL);
	write_register(0, UNIV_LSI3_CTL);
	write_register(0, UNIV_VSI0_CTL);
	write_register(0, UNIV_VSI1_CTL);
	write_register(0, UNIV_VSI2_CTL);
	write_register(0, UNIV_VSI3_CTL);
      }
    if (!found)
      {
	printk("universe: No Universe chip found. giving up.\n");
	unregister_chrdev(UNIVERSE_MAJOR,"universe");
	return -ENOSYS;
      }
  }
  
  {
    int ret;
    int irq;
    irq = read_register(UNIV_PCI_MISC1) & 0x000000FF;
    ret = request_irq(irq, universe_interrupt, 0, "universe", NULL);
    
    /* map all interrupt sources to pin 0 */
    write_register(0,UNIV_LINT_MAP0);
    write_register(0,UNIV_LINT_MAP1);
    /* enable most interrupts */
    write_register(/* UNIV_BM_LINT_ACFAIL | UNIV_BM_LINT_SYSFAIL | */
		   /* UNIV_BM_LINT_SW_INT | */ UNIV_BM_LINT_SW_IACK |
		   UNIV_BM_LINT_VERR | UNIV_BM_LINT_LERR | UNIV_BM_LINT_DMA |
		   UNIV_BM_LINT_VIRQ | UNIV_BM_LINT_VOWN,
		   UNIV_LINT_EN);
    /* set latency timer */
    write_register(32 << UNIV_OF_PCI_MISC0_LTIMER,UNIV_PCI_MISC0);
    /* set PCI retries */
    {
      /* set PCI retries to 0 (Disabled, i.e. infinite number allowed,
	 because of a bug in the CA91C042 silicon. See Universe (CA91C042)
	 Errata manual, erratum 6, Local Master MAXRTRY failure.
	 With the new revision (Universe II) the retry counter may be set
	 to a nonzero value... */
      unsigned int mast_ctl;
      mast_ctl = read_register( UNIV_MAST_CTL);
      mast_ctl &= ~UNIV_BM_MAST_CTL_MAXRTRY;
      mast_ctl |= UNIV_BM_MAST_CTL_MAXRTRY & 
	(0 << UNIV_OF_MAST_CTL_MAXRTRY);
      /* set PCI Aligned burst size to 32 byte,
	 as recomended by Errata manual. erratum 14: DMA Write cycles
	 with PCI aligned burst size - 64 Bytes */
      mast_ctl &= ~UNIV_BM_MAST_CTL_PABS;
      write_register(mast_ctl,UNIV_MAST_CTL);
    }
    /* set VME timeout */
    {
      unsigned int misc_ctl;
      misc_ctl = read_register( UNIV_MISC_CTL);
      misc_ctl &= ~UNIV_BM_MISC_CTL_VBTO;
      misc_ctl |= UNIV_BM_MISC_CTL_VBTO & 
	((0x7) << UNIV_OF_MISC_CTL_VBTO);
      write_register(misc_ctl,UNIV_MISC_CTL);
    }
    { /* clear the SYSFAIL line, which is active by default */
      unsigned int vcsr;
      vcsr = read_register( UNIV_VCSR_CLR);
      printk("universe: VCSR is %08x \n",vcsr);
      vcsr = UNIV_BM_VCSR_SYSFAIL;
      write_register( vcsr, UNIV_VCSR_CLR);
    }
    { /* reset all interrupt bits that are on, otherwise interrupts might
	 not be really enabled, see also interrupt service routine  */
      unsigned int lint_stat;
      lint_stat = read_register( UNIV_LINT_STAT);
      write_register (lint_stat, UNIV_LINT_STAT);
    }
    { /* PCI bus mastering must be enabled or DMA will not work,
	 as well as VME slave windows, which also require PCI Bus mastership */
      unsigned short pci_csr, pci_csr_orig;
      printk("universe: will enable PCI Bus master\n");
      pci_csr_orig = read_register_word(PCI_COMMAND);
      pci_csr = pci_csr_orig | PCI_COMMAND_MASTER;
      write_register_word(pci_csr, PCI_COMMAND);
      printk("universe: PCI Bus master enabled\n");
    }
  }
  printk(KERN_INFO "universe driver $Id: universe.c,v 1.43 1997/06/09 10:46:00 hannappe Exp $ installed, major %d\n",UNIVERSE_MAJOR);
  printk(KERN_INFO "universe driver compiled " __DATE__ " " __TIME__ "\n");
  return(0);
}

void cleanup_module(void)
{
  if (Status_buffer)
    {
      printk(KERN_ERR "universe: Status buffer left over, deleting it\n");
      kfree(Status_buffer);
      Status_buffer = NULL;
    }
  if (Slsi_kernel_addr)
    {
      printk(KERN_INFO "universe: SLSI still mapped, unmapping\n");
      iounmap(Slsi_kernel_addr);
      Slsi_kernel_addr = NULL;
    }
  {
    int irq;
    irq = read_register(UNIV_PCI_MISC1) & 0x000000FF;
    printk(KERN_INFO "universe: will now free irq %d\n",irq);
    write_register(0,UNIV_LINT_EN); /* disable all pci interrupts */
    free_irq(irq,NULL);
  }
#ifdef UNIVERSE_REGISTERS_ARE_MEMORY_MAPPED
  iounmap(UCSR_base_addr);
#endif
  unregister_chrdev(UNIVERSE_MAJOR,"universe");
  printk(KERN_INFO "universe driver removed\n");
}
